from ..registry import register_model  # noqa
from .model import Model, ModelConfig  # noqa
from .audio_classification import *
from .backbone import *
from .image2text import *
from .language_modeling import *
from .sequence_labeling import *
from .speech_recognition import *
from .text_classification import *
from .text_detection import *
from .text_embedding import *
from .text_generation import *
