from ..registry import register_embedding  # noqa
from .embedding import Embedding, EmbeddingConfig  # noqa
from .fasttext import FastText, FastTextConfig
from .word2vec import Word2Vec, Word2VecConfig
