from .beit_roberta import BeitRobertaImage2Text, BeitRobertaImage2TextConfig
from .crnn import CRNNImage2Text, CRNNImage2TextConfig
from .trocr import TrOCRImage2Text, TrOCRImage2TextConfig
from .vit_gpt2 import ViTGPT2Image2Text, ViTGPT2Image2TextConfig
from .vit_roberta import ViTRobertaImage2Text, ViTRobertaImage2TextConfig
